# Scola version 1.1.0
# ScolaMetaModel.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported Modules
# 2) Elements
# 3) Containers
# 4) Internal Exceptions

# 1) Imported Modules
# -------------------

import string
import sys
import re

from ScolaPaths import *

# 2) Elements
# -----------

class ScolaElement:
  def __init__(self, type, container):
    self._isAtomic = True
    self._type = type
    self._identifier = None
    self._path = None
    self._container = container
    if container!=None:
      container._AddElement(self)
    self._isContainer = False
    self._nextElement = None
    self._systemPath = None
    self._flag = None

  def Delete(self):
    if self._container==None:
      return
    if self._identifier!=None:
      self._container._RemoveFromDictonary(self)
    self._container._RemoveElement(self)

  def GetType(self):
    return self._type

  def GetIdentifier(self):
    return self._identifier
  def SetIdentifier(self, identifier):
    self._identifier = identifier
  def Rename(self, identifier):
    if self._container==None:
      self._identifier = identifier
      return
    if self._identifier!=None:
      self._container._RemoveFromDictonary(self)
    self._identifier = identifier
    self._container._AddToDictonary(self)

  def GetPath(self):
    return self._path
  def SetPath(self, path):
    self._path = path

  def GetAbsolutePath(self):
    if self.GetIdentifier()==None:
      return None
    path = ScolaPath()
    path.PrependIdentifier(self.GetIdentifier())
    container = self.GetContainer()
    if container!=None:
      while container.GetContainer()!=None:
        path.PrependIdentifier(container.GetIdentifier())
        container = container.GetContainer()
    return path

  def GetContainer(self):
    return self._container
  def GetModel(self):
    if self.GetContainer()==None:
      return self
    return self.GetContainer().GetModel()

  def GetNextElement(self):
    return self._nextElement
  def SetNextElement(self, element):
    self._nextElement = element

  def GetSystemPath(self):
    return self._systemPath
  def SetSystemPath(self, path):
    self._systemPath = path

  def GetFlag(self):
    return self._flag
  def SetFlag(self, value):
    self._flag = value

# 3) Containers
# -------------

class ScolaContainer(ScolaElement):
  def __init__(self, type, container):
    ScolaElement.__init__(self, type, container)
    self._isAtomic = False
    self._isContainer = True
    self._elements = []
    self._dictionary = dict()

  def DeleteAll(self):
    while self.GetNumberOfElements()>0:
      element = self.GetElement(0)
      element.Delete()
    self.Delete()

  def GetNumberOfElements(self):
    return len(self._elements)
  def GetElement(self, index):
    if index<0 or index>=self.GetNumberOfElements():
      return None
    return self._elements[index]
  def GetElements(self):
    return self._elements
  def GetDictonary(self):
    return self._dictionary
  def LookForNamedElement(self, identifier):
    return self._dictionary.get(identifier, None)
  def _AddElement(self, element):
    self._elements.append(element)
  def _RemoveElement(self, element):
    self._elements.remove(element)
  def _AddToDictonary(self, element):
    # REQUIRE: self.LookForNamedElement(element.GetIdentifier())==None
    self._dictionary[element.GetIdentifier()] = element
  def _RemoveFromDictonary(self, element):
    del self._dictionary[element.GetIdentifier()]
  def AccessNamedElement(self, path):
    if path.IsEmpty():
      return self
    identifier = path.RemoveFirstIdentifier()
    if identifier==ScolaPath.ROOT_IDENTIFIER:
      container = self.GetModel()
      element = container.AccessNamedElement(path)
    elif identifier==ScolaPath.CONTAINER_IDENTIFIER:
      container = self.GetContainer()
      if container==None:
        element = None
      else:
        element = container.AccessNamedElement(path)
    else:
      element = self.LookForNamedElement(identifier)
      if element!=None:
        if element._isContainer:
          container = element
          element = container.AccessNamedElement(path)
        elif not path.IsEmpty():
          element = None
    path.PrependIdentifier(identifier)
    return element
  def RemoveElement(self, element):
    if not element in self._elements:
      return -1
    index = self._elements.index(element)
    self._elements.remove(element)
    if element.GetIdentifier()!=None:
      self._RemoveFromDictonary(element)
    element._container = None
    return index
  def AddElement(self, element):
    self._elements.append(element)
    if element.GetIdentifier()!=None:
      self._AddToDictonary(element)
    element._container = self
  def InsertElement(self, index, element):
    self._elements.insert(index, element)
    if element.GetIdentifier()!=None:
      self._AddToDictonary(element)
    element._container = self

  def CollectElements(self, test, elements):
    for element in self.GetElements():
      if element._isAtomic:
        if test(element):
          elements.append(element)
      else:
        element.CollectElements(test, elements)
    if test(self):
      elements.append(self)

# 4) Internal Exceptions
# ----------------------

class ScolaInternalError(Exception):
  def __init__(self, message):
    self._message = message
