# Scola version 1.1.0
# ScolaLexer.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Token
# 3) Lexer

# 1) Imported modules
# -------------------

import string
import sys
import re

# 2) Token
# --------

class ScolaToken:
  LPAR = "LPAR"
  RPAR = "RPAR"
  LBRACE = "LBRACE"
  RBRACE = "RBRACE"
  COMMA = "COMMA"
  QUOTE = "QUOTE"
  INTEGER = "INTEGER"
  REAL = "REAL"
  IDENTIFIER = "IDENTIFIER"
  PATH = "PATH"
  STRING = "STRING"
  KEYWORD_CLASS = "KEYWORD_CLASS"
  KEYWORD_CLONES = "KEYWORD_CLONES"
  KEYWORD_INHERITS = "KEYWORD_INHERITS"
  KEYWORD_DOMAIN = "KEYWORD_DOMAIN"
  KEYWORD_BLOCK = "KEYWORD_BLOCK"
  KEYWORD_PORT = "KEYWORD_PORT"
  KEYWORD_ASSERTION = "KEYWORD_ASSERTION"
  KEYWORD_SCENARIO = "KEYWORD_SCENARIO"
  KEYWORD_AS = "KEYWORD_AS"
  KEYWORD_STATE = "KEYWORD_STATE"
  KEYWORD_TASK = "KEYWORD_TASK"
  KEYWORD_TEST = "KEYWORD_TEST"
  KEYWORD_CASE = "KEYWORD_CASE"
  KEYWORD_CHOICE = "KEYWORD_CHOICE"
  KEYWORD_BRANCH = "KEYWORD_BRANCH"
  KEYWORD_FORK = "KEYWORD_FORK"
  KEYWORD_JOIN = "KEYWORD_JOIN"
  KEYWORD_SPLIT = "KEYWORD_SPLIT"
  KEYWORD_MERGE = "KEYWORD_MERGE"
  KEYWORD_MEET = "KEYWORD_MEET"
  KEYWORD_END = "KEYWORD_END"
  KEYWORD_NEXT = "KEYWORD_NEXT"
  KEYWORD_FAIL = "KEYWORD_FAIL"
  KEYWORD_SET = "KEYWORD_SET"
  KEYWORD_NEW = "KEYWORD_NEW"
  KEYWORD_DELETE = "KEYWORD_DELETE"
  KEYWORD_MOVE = "KEYWORD_MOVE"
  KEYWORD_SEND = "KEYWORD_SEND"
  KEYWORD_RECEIVE = "KEYWORD_RECEIVE"
  KEYWORD_CLONE = "KEYWORD_CLONE"
  KEYWORD_THEN = "KEYWORD_THEN"
  KEYWORD_ELSE = "KEYWORD_ELSE"
  KEYWORD_IF = "KEYWORD_IF"
  KEYWORD_BEGIN = "KEYWORD_BEGIN"
  KEYWORD_PRINT = "KEYWORD_PRINT"
  KEYWORD_NONE = "KEYWORD_NONE"
  KEYWORD_TRUE = "KEYWORD_TRUE"
  KEYWORD_FALSE = "KEYWORD_FALSE"
  KEYWORD_AND = "KEYWORD_AND"
  KEYWORD_OR = "KEYWORD_OR"
  KEYWORD_NOT = "KEYWORD_NOT"
  KEYWORD_EQ = "KEYWORD_EQ"
  KEYWORD_DF = "KEYWORD_DF"
  KEYWORD_LT = "KEYWORD_LT"
  KEYWORD_GT = "KEYWORD_GT"
  KEYWORD_LEQ = "KEYWORD_LEQ"
  KEYWORD_GEQ = "KEYWORD_GEQ"
  KEYWORD_OPP = "KEYWORD_OPP"
  KEYWORD_INV = "KEYWORD_INV"
  KEYWORD_ADD = "KEYWORD_ADD"
  KEYWORD_SUB = "KEYWORD_SUB"
  KEYWORD_MUL = "KEYWORD_MUL"
  KEYWORD_DIV = "KEYWORD_DIV"
  KEYWORD_MIN = "KEYWORD_MIN"
  KEYWORD_MAX = "KEYWORD_MAX"
  KEYWORD_ABS = "KEYWORD_ABS"
  KEYWORD_POW = "KEYWORD_POW"
  KEYWORD_EXP = "KEYWORD_EXP"
  KEYWORD_LOG = "KEYWORD_LOG"
  KEYWORD_SQRT = "KEYWORD_SQRT"
  KEYWORD_MOD = "KEYWORD_MOD"
  KEYWORD_CEIL = "KEYWORD_CEIL"
  KEYWORD_FLOOR = "KEYWORD_FLOOR"
  KEYWORD_COUNT = "KEYWORD_COUNT"
  KEYWORD_APPEND = "KEYWORD_APPEND"
  KEYWORD_BOOLEAN = "KEYWORD_BOOLEAN"
  KEYWORD_INTEGER = "KEYWORD_INTEGER"
  KEYWORD_REAL = "KEYWORD_REAL"
  KEYWORD_SYMBOL = "KEYWORD_SYMBOL"
  KEYWORD_IDENTIFIER = "KEYWORD_IDENTIFIER"
  KEYWORD_OWNER = "KEYWORD_OWNER"
  KEYWORD_STRING = "KEYWORD_STRING"
  KEYWORD_IS_PORT = "KEYWORD_IS_PORT"
  KEYWORD_IS_BLOCK = "KEYWORD_IS_BLOCK"
  KEYWORD_IS_ASSERTION = "KEYWORD_IS_ASSERTION"
  KEYWORD_SIZE = "KEYWORD_SIZE"
  KEYWORD_ELEMENT = "KEYWORD_ELEMENT"
  KEYWORD_QUOTE = "KEYWORD_QUOTE"
  KEYWORD_EVAL = "KEYWORD_EVAL"
  COMMENT = "COMMENT"
  UNKNOWN = "UNKNOWN"
  NON_TERMINATED_STRING = "NON_TERMINATED_STRING"
  NON_TERMINATED_COMMENT = "NON_TERMINATED_COMMENT"

  def __init__(self, type, string, fileName, lineNumber, columnNumber):
    self.type = type
    self.string = string
    self.fileName = fileName
    self.lineNumber = lineNumber
    self.columnNumber = columnNumber

  @staticmethod
  def GetNameOfType(type):
    return type

# 3) Lexer
# --------

class ScolaLexer:
  _Keywords = {
    # Systems
    "clones": ScolaToken.KEYWORD_CLONES,
    "class": ScolaToken.KEYWORD_CLASS,
    "inherits": ScolaToken.KEYWORD_INHERITS,
    "domain": ScolaToken.KEYWORD_DOMAIN,
    "port": ScolaToken.KEYWORD_PORT,
    "block": ScolaToken.KEYWORD_BLOCK,
    "assertion": ScolaToken.KEYWORD_ASSERTION,
    # Use cases
    "scenario": ScolaToken.KEYWORD_SCENARIO,
    "as": ScolaToken.KEYWORD_AS,
    "state": ScolaToken.KEYWORD_STATE,
    "task": ScolaToken.KEYWORD_TASK,
    "test": ScolaToken.KEYWORD_TEST,
    "case": ScolaToken.KEYWORD_CASE,
    "choice": ScolaToken.KEYWORD_CHOICE,
    "branch": ScolaToken.KEYWORD_BRANCH,
    "fork": ScolaToken.KEYWORD_FORK,
    "join": ScolaToken.KEYWORD_JOIN,
    "split": ScolaToken.KEYWORD_SPLIT,
    "merge": ScolaToken.KEYWORD_MERGE,
    "meet": ScolaToken.KEYWORD_MEET,
    "end": ScolaToken.KEYWORD_END,
    "next": ScolaToken.KEYWORD_NEXT,
    # Instructions
    "fail": ScolaToken.KEYWORD_FAIL,
    "set": ScolaToken.KEYWORD_SET,
    "new": ScolaToken.KEYWORD_NEW,
    "delete": ScolaToken.KEYWORD_DELETE,
    "move": ScolaToken.KEYWORD_MOVE,
    "send": ScolaToken.KEYWORD_SEND,
    "receive": ScolaToken.KEYWORD_RECEIVE,
    "clone": ScolaToken.KEYWORD_CLONE,
    "then": ScolaToken.KEYWORD_THEN,
    "else": ScolaToken.KEYWORD_ELSE,
    "begin": ScolaToken.KEYWORD_BEGIN,
    "print": ScolaToken.KEYWORD_PRINT,
    # Expressions
    "none": ScolaToken.KEYWORD_NONE,
    "true": ScolaToken.KEYWORD_TRUE,
    "false": ScolaToken.KEYWORD_FALSE,
    "and": ScolaToken.KEYWORD_AND,
    "or": ScolaToken.KEYWORD_OR,
    "not": ScolaToken.KEYWORD_NOT,
    "eq": ScolaToken.KEYWORD_EQ,
    "df": ScolaToken.KEYWORD_DF,
    "lt": ScolaToken.KEYWORD_LT,
    "gt": ScolaToken.KEYWORD_GT,
    "leq": ScolaToken.KEYWORD_LEQ,
    "geq": ScolaToken.KEYWORD_GEQ,
    "opp": ScolaToken.KEYWORD_OPP,
    "inv": ScolaToken.KEYWORD_INV,
    "add": ScolaToken.KEYWORD_ADD,
    "sub": ScolaToken.KEYWORD_SUB,
    "mul": ScolaToken.KEYWORD_MUL,
    "div": ScolaToken.KEYWORD_DIV,
    "min": ScolaToken.KEYWORD_MIN,
    "max": ScolaToken.KEYWORD_MAX,
    "abs": ScolaToken.KEYWORD_ABS,
    "pow": ScolaToken.KEYWORD_POW,
    "exp": ScolaToken.KEYWORD_EXP,
    "log": ScolaToken.KEYWORD_LOG,
    "sqrt": ScolaToken.KEYWORD_SQRT,
    "mod": ScolaToken.KEYWORD_MOD,
    "ceil": ScolaToken.KEYWORD_CEIL,
    "floor": ScolaToken.KEYWORD_FLOOR,
    "count": ScolaToken.KEYWORD_COUNT,
    "append": ScolaToken.KEYWORD_APPEND,
    "if": ScolaToken.KEYWORD_IF,
    "Boolean": ScolaToken.KEYWORD_BOOLEAN,
    "integer": ScolaToken.KEYWORD_INTEGER,
    "real": ScolaToken.KEYWORD_REAL,
    "symbol": ScolaToken.KEYWORD_SYMBOL,
    "owner": ScolaToken.KEYWORD_OWNER,
    "identifier": ScolaToken.KEYWORD_IDENTIFIER,
    "owner": ScolaToken.KEYWORD_OWNER,
    "string": ScolaToken.KEYWORD_STRING,
    "is_block": ScolaToken.KEYWORD_IS_BLOCK,
    "is_port": ScolaToken.KEYWORD_IS_PORT,
    "is_assertion": ScolaToken.KEYWORD_IS_ASSERTION,
    "size": ScolaToken.KEYWORD_SIZE,
    "element": ScolaToken.KEYWORD_ELEMENT,
    "quote": ScolaToken.KEYWORD_QUOTE,
    "eval": ScolaToken.KEYWORD_EVAL
    }
  END_OF_FILE = 0
  INITIAL = 1
  IDENTIFIER = 10
  PATH_1 = 11
  PATH_2 = 12
  NUMBER_1 = 20
  NUMBER_2 = 21
  NUMBER_3 = 22
  NUMBER_4 = 23
  NUMBER_5 = 24
  NUMBER_6 = 25
  NUMBER_7 = 26
  NUMBER_8 = 27
  STRING_1 = 30
  STRING_2 = 31
  SINGLE_LINE_COMMENT = 40
  MULTILINE_COMMENT_1 = 50
  MULTILINE_COMMENT_2 = 51

  def __init__(self):
    self.keepComments = False
    self.fileName = None
    self.file = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.line = None
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.tokens = []
    self.state = -1

  def Initialize(self, fileName, file):
    self.fileName = fileName
    self.file = file
    self.tokens = []
    self.line = None
    self.lineNumber = 0
    self.columnNumber = 0
    self.endOfFile = 0
    self.tokenLineNumber = 0
    self.tokenColumnNumber = 0
    self.tokenString = None
    self.readChars = []
    self.GetNewLine()
    self.state = ScolaLexer.INITIAL
    return 0

  def Finalize(self):
    pass
     # self.file.close()

  def GetToken(self, tokenIndex):
    if tokenIndex<len(self.tokens):
      return self.tokens[tokenIndex]
    while self.state!=ScolaLexer.END_OF_FILE and tokenIndex>=len(self.tokens):
      if self.state==ScolaLexer.INITIAL:
        c = self.GetChar()
        self.tokenLineNumber = self.lineNumber
        self.tokenColumnNumber = self.columnNumber-1
        if c==None:
          self.state = ScolaLexer.END_OF_FILE
        elif c.isspace():
          continue
        elif c=="(":
          self.tokenString = c
          self.NewToken(ScolaToken.LPAR)
        elif c==")":
          self.tokenString = c
          self.NewToken(ScolaToken.RPAR)
        elif c=="{":
          self.tokenString = c
          self.NewToken(ScolaToken.LBRACE)
        elif c=="}":
          self.tokenString = c
          self.NewToken(ScolaToken.RBRACE)
        elif c==",":
          self.tokenString = c
          self.NewToken(ScolaToken.COMMA)
        elif c=="+" or c=="-":
          c = self.GetChar()
          if c.isdigit():
            self.tokenString = "+" + c
            self.state = ScolaLexer.NUMBER_1
          else:
            self.tokenString = "+"
            self.PutCharBack(c)
            self.NewToken(ScolaToken.UNKNOWN)
        elif c=="'":
          self.tokenString = c
          self.NewToken(ScolaToken.QUOTE)
        elif c=="/":
          c = self.GetChar()
          if c=="/":
            self.tokenString = "//"
            self.state = ScolaLexer.SINGLE_LINE_COMMENT
          elif c=="*":
            self.tokenString = "/*"
            self.state = ScolaLexer.MULTILINE_COMMENT_1
          else:
            self.tokenString = "/"
            self.PutCharBack(c)
            self.NewToken(ScolaToken.UNKNOWN)
        elif c.isalpha() or c=="_":
          self.tokenString = c
          self.state = ScolaLexer.IDENTIFIER
        elif c.isdigit():
          self.tokenString = c
          self.state = ScolaLexer.NUMBER_2
        elif c=='"':
          # self.tokenString = c
          self.tokenString = ""
          self.state = ScolaLexer.STRING_1
        else:
          self.tokenString = c
          self.NewToken(ScolaToken.UNKNOWN)

      # -----------------------------------------------------
      # Identifiers & Paths
      # -----------------------------------------------------
      elif self.state==ScolaLexer.IDENTIFIER:
        c = self.GetChar()
        if c.isdigit() or c.isalpha() or c=="_":
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = ScolaLexer.PATH_1
        else:
          self.PutCharBack(c)
          type = self.GetIdentifierType(self.tokenString)
          self.NewToken(type)
          self.state = ScolaLexer.INITIAL

      elif self.state==ScolaLexer.PATH_1:
        c = self.GetChar()
        if c.isalpha() or c=="_":
          self.tokenString += c
          self.state = ScolaLexer.PATH_2
        else:
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.PATH_2:
        c = self.GetChar()
        if c.isdigit() or c.isalpha() or c=="_":
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = ScolaLexer.PATH_1
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.PATH)
          self.state = ScolaLexer.INITIAL

      # -----------------------------------------------------
      # Numbers
      # -----------------------------------------------------
      elif self.state==ScolaLexer.NUMBER_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.UNKNOW)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_2
        elif c==".":
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_8
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.UNKNOW)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.NUMBER_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.INTEGER)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        elif c==".":
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_3
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.INTEGER)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.NUMBER_3:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_4
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.NUMBER_4:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.REAL)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        elif c=="e" or c=="E":
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_5
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.REAL)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.NUMBER_5:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_7
        elif c=="+" or c=="-":
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_6
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.NUMBER_6:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_7
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.NUMBER_7:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.REAL)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.REAL)
          self.state = ScolaLexer.INITIAL
      elif self.state==ScolaLexer.NUMBER_8:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.END_OF_FILE
        elif c.isdigit():
          self.tokenString += c
          self.state = ScolaLexer.NUMBER_4
        else:
          self.PutCharBack(c)
          self.NewToken(ScolaToken.UNKNOWN)
          self.state = ScolaLexer.INITIAL

      # -----------------------------------------------------
      # Strings
      # -----------------------------------------------------
      elif self.state==ScolaLexer.STRING_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.NON_TERMINATED_STRING)
          self.state = ScolaLexer.END_OF_FILE
        elif c=='"':
          # self.tokenString += c
          self.NewToken(ScolaToken.STRING)
          self.state = ScolaLexer.INITIAL
        elif c=='\\':
          self.tokenString += c
          c = self.GetChar()
          self.state = ScolaLexer.STRING_2
        else:
          self.tokenString += c
      elif self.state==ScolaLexer.STRING_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.NON_TERMINATED_STRING)
          self.state = ScolaLexer.END_OF_FILE
        elif c=='\\':
          self.tokenString += c
        else:
          self.tokenString += c
          self.state = ScolaLexer.STRING_1

      # -----------------------------------------------------
      # Single line comments
      # -----------------------------------------------------
      elif self.state==ScolaLexer.SINGLE_LINE_COMMENT:
        c = self.GetChar()
        if c==None:
          if self.keepComments:
            self.NewToken(ScolaToken.COMMENT)
          self.state = 0
        elif c=="\n":
          if self.keepComments:
            self.NewToken(ScolaToken.COMMENT)
          self.state = 1
        else:
          self.tokenString += c

      # -----------------------------------------------------
      # Multiline comments
      # -----------------------------------------------------
      elif self.state==ScolaLexer.MULTILINE_COMMENT_1:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.NON_TERMINATED_COMMENT)
          self.state = ScolaLexer.END_OF_FILE
        elif c=="*":
          self.tokenString += c
          self.state = ScolaLexer.MULTILINE_COMMENT_2
        else:
          self.tokenString += c
      elif self.state==ScolaLexer.MULTILINE_COMMENT_2:
        c = self.GetChar()
        if c==None:
          self.NewToken(ScolaToken.NON_TERMINATED_COMMENT)
          self.state = ScolaLexer.END_OF_FILE
        elif c=="*":
          self.tokenString += c
          c = self.GetChar()
        elif c=="/":
          self.tokenString += c
          if self.keepComments:
            self.NewToken(ScolaToken.COMMENT)
          self.state = ScolaLexer.INITIAL
        else:
          self.tokenString += c
          self.state = ScolaLexer.MULTILINE_COMMENT_1

    if tokenIndex>=len(self.tokens):
      return None
    return self.tokens[tokenIndex]

  def GetIdentifierType(self, identifier):
    return ScolaLexer._Keywords.get(identifier, ScolaToken.IDENTIFIER)

  def NewToken(self, type):
    token = ScolaToken(type, self.tokenString, self.fileName, self.tokenLineNumber, self.tokenColumnNumber)
    self.tokens.append(token)

  def GetChar(self):
    if len(self.readChars)!=0:
      return self.readChars.pop()
    if self.columnNumber==len(self.line):
      self.GetNewLine()
      if self.endOfFile:
        return None
      return self.GetChar()
    c = self.line[self.columnNumber]
    self.columnNumber += 1
    return c

  def PutCharBack(self, c):
    self.readChars.append(c)

  def GetNewLine(self):
    self.line = self.file.readline()
    if self.line=="":
      self.endOfFile = 1
      return
    self.lineNumber += 1
    self.columnNumber = 0
    self.line = self.line.rstrip()
    self.line += "\n"

  def ScanFile(self, fileName):
    error = self.Initialize(fileName)
    if error!=0:
      return
    tokenIndex = 0
    while True:
      token = self.GetToken(tokenIndex)
      if token==None:
        break
      tokenIndex += 1
    self.Finalize()

  def PrintTokens(self, outputFile):
    for token in self.tokens:
      outputFile.write(str(token.lineNumber) + "\t")
      outputFile.write(str(token.columnNumber) + "\t")
      outputFile.write(ScolaToken.GetNameOfType(token.type) + "\t")
      outputFile.write(token.string + "\n")
